from django.apps import AppConfig


class TestsConfig(AppConfig):
    name = 'pgpubsub.tests'
    default_auto_field = 'django.db.models.AutoField'

    def ready(self):
        import pgpubsub.tests.listeners  # noqa



# Generated by Django 3.2.12 on 2023-05-30 10:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0009_auto_20230522_0720'),
    ]

    operations = [
        migrations.AddField(
            model_name='media',
            name='store_id',
            field=models.TextField(null=True),
        ),
    ]



# Generated by Django 3.2.12 on 2023-05-03 06:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pgpubsub', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, null=True),
        ),
    ]



from typing import Type

from django.db import models

try:
    from django.db.models import JSONField
except:
    from django.contrib.postgres.fields import JSONField

from pgpubsub.channel import BaseChannel
import pgtrigger


MAX_POSTGRES_CHANNEL_LENGTH = 63


class Notification(models.Model):
    channel = models.CharField(max_length=MAX_POSTGRES_CHANNEL_LENGTH)
    payload = JSONField()
    # The field is made nullable to make sure the addiion of the new field is backward
    # compatible. From the version this change is release the field is effectively non
    # nullable as in it always gets a value.
    # After some time the field should be made non nullable here.
    created_at = models.DateTimeField(auto_now_add=True, null=True)
    # This is a FK for the migration id but ForeignKey cannot be used as migrations
    # is not a regular table in django, so FK is created manually.
    db_version = models.IntegerField(null=True)

    class Meta:
        triggers = [
            pgtrigger.Trigger(
                name="pgpubsub_notification_set_db_version",
                operation=pgtrigger.Insert,
                when=pgtrigger.Before,
                func="""
                    NEW.db_version := (
                        SELECT max(id)
                        FROM django_migrations
                        WHERE app = ((NEW.payload #>> '{}')::jsonb ->> 'app')
                    );
                    NEW.created_at := NOW();
                    RETURN NEW;
                """,
            )
        ]
        ordering = ['created_at']

    def __repr__(self):
        return (
            f'Notification('
            f'channel={self.channel},'
            f' payload={self.payload},'
            f' created_at={self.created_at},'
            f' db_version={self.db_version}'
            f')'
        )

    @classmethod
    def from_channel(cls, channel: Type[BaseChannel]):
        return cls.objects.filter(channel=channel.listen_safe_name())



# Generated by Django 3.2.22 on 2023-10-20 00:17

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pgpubsub', '0004_notification_pgpubsub_notification_set_db_version'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='notification',
            options={'ordering': ['created_at']},
        ),
    ]



# Generated by Django 3.2.12 on 2023-05-16 11:51

from django.db import migrations
import pgtrigger.compiler
import pgtrigger.migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0002_auto_20230411_0829'),
    ]

    operations = [
        pgtrigger.migrations.AddTrigger(
            model_name='author',
            trigger=pgtrigger.compiler.Trigger(name='pgpubsub_160cf', sql=pgtrigger.compiler.UpsertTriggerSql(declare='DECLARE payload TEXT;', func="\n            \n            payload := json_build_object(\n                'app', 'tests',\n                'model', 'Author',\n                'old', row_to_json(OLD),\n                'new', row_to_json(NEW)\n              );\n        \n            \n            INSERT INTO pgpubsub_notification (channel, payload)\n            VALUES ('pgpubsub_160cf', to_json(payload::text));\n        \n            perform pg_notify('pgpubsub_160cf', payload);\n            RETURN NEW;\n        ", hash='200718f1bddbea02e3c41f60ccea5676f0ae8860', operation='INSERT', pgid='pgtrigger_pgpubsub_160cf_72a36', table='tests_author', when='AFTER')),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='media',
            trigger=pgtrigger.compiler.Trigger(name='pgpubsub_a83de', sql=pgtrigger.compiler.UpsertTriggerSql(declare='DECLARE payload TEXT;', func="\n            \n            payload := json_build_object(\n                'app', 'tests',\n                'model', 'Media',\n                'old', row_to_json(OLD),\n                'new', row_to_json(NEW)\n              );\n        \n            \n            INSERT INTO pgpubsub_notification (channel, payload)\n            VALUES ('pgpubsub_a83de', to_json(payload::text));\n        \n            perform pg_notify('pgpubsub_a83de', payload);\n            RETURN NEW;\n        ", hash='658278a249af2641e4e16deb18555b35ab20e597', operation='UPDATE OR INSERT', pgid='pgtrigger_pgpubsub_a83de_cacbb', table='tests_media', when='AFTER')),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='post',
            trigger=pgtrigger.compiler.Trigger(name='pgpubsub_72091', sql=pgtrigger.compiler.UpsertTriggerSql(declare='DECLARE payload TEXT;', func="\n            \n            payload := json_build_object(\n                'app', 'tests',\n                'model', 'Post',\n                'old', row_to_json(OLD),\n                'new', row_to_json(NEW)\n              );\n        \n            \n            INSERT INTO pgpubsub_notification (channel, payload)\n            VALUES ('pgpubsub_72091', to_json(payload::text));\n        \n            perform pg_notify('pgpubsub_72091', payload);\n            RETURN NEW;\n        ", hash='16a6a1629e6a9fe42f73d460319a92689efdcbfe', operation='DELETE', pgid='pgtrigger_pgpubsub_72091_67aeb', table='tests_post', when='AFTER')),
        ),
    ]



import datetime
import json
from unittest.mock import patch

from django.db import connection
from django.db.transaction import atomic
from django.db.migrations.recorder import MigrationRecorder
import pytest

from pgpubsub.listen import (
    listen_to_channels,
    process_notifications,
    listen,
)
from pgpubsub.models import Notification
from pgpubsub.notify import process_stored_notifications
from pgpubsub.tests.channels import (
    AuthorTriggerChannel,
    MediaTriggerChannel,
)
from pgpubsub.tests.listeners import post_reads_per_date_cache
from pgpubsub.tests.models import Author, Media, Post


@pytest.fixture()
def pg_connection():
    return listen_to_channels()


@pytest.mark.django_db(transaction=True)
def test_post_fetch_notify(pg_connection):
    author = Author.objects.create(name='Billy')
    Notification.from_channel(channel=AuthorTriggerChannel).get()
    assert 1 == len(pg_connection.notifies)
    today = datetime.date.today()
    post = Post.objects.create(
        author=author, content='first post', date=today)
    assert post_reads_per_date_cache[today] == {}
    Post.fetch(post.pk)
    assert 1 == Notification.objects.count()
    pg_connection.poll()
    assert 2 == len(pg_connection.notifies)
    process_notifications(pg_connection)
    assert post_reads_per_date_cache[today] == {post.pk: 1}
    assert 0 == Notification.objects.count()


@pytest.mark.django_db(transaction=True)
def test_author_insert_notify(pg_connection):
    author = Author.objects.create(name='Billy')
    assert 1 == len(pg_connection.notifies)
    stored_notification = Notification.from_channel(
        channel=AuthorTriggerChannel).get()
    assert 'old' in stored_notification.payload
    assert 'new' in stored_notification.payload
    assert not Post.objects.exists()
    process_notifications(pg_connection)
    assert 1 == Post.objects.count()
    post = Post.objects.last()
    assert post.author == author


@pytest.mark.django_db(transaction=True)
def test_author_insert_notify_in_transaction(pg_connection):
    with atomic():
        author = Author.objects.create(name='Billy')
    pg_connection.poll()
    assert 1 == len(pg_connection.notifies)
    assert not Post.objects.exists()
    process_notifications(pg_connection)
    assert 1 == Post.objects.count()
    post = Post.objects.last()
    assert post.author == author


@pytest.mark.django_db(transaction=True)
def test_author_insert_notify_transaction_rollback(pg_connection):
    class TestException(Exception):
        pass

    try:
        with atomic():
            Author.objects.create(name='Billy')
            raise TestException
    except TestException:
        pass

    # Notifications are only sent when the transaction commits
    pg_connection.poll()
    assert not pg_connection.notifies
    assert not Author.objects.exists()
    assert not Post.objects.exists()


@pytest.mark.django_db(transaction=True)
def test_author_bulk_insert_notify(pg_connection):
    authors = [Author(name='Billy'), Author(name='Craig')]
    with atomic():
        authors = Author.objects.bulk_create(authors)

    # TODO: Understand why pg_connection.poll() is only
    # necessary when we invoke a notification inside
    # a transaction (which happens in a bulk_create).
    pg_connection.poll()
    assert 2 == len(pg_connection.notifies)
    assert not Post.objects.exists()
    process_notifications(pg_connection)
    assert 2 == Post.objects.count()
    post_authors = Post.objects.values_list('author_id', flat=True)
    assert [author.pk for author in authors] == list(post_authors)


def _simulate_listener_does_not_receive_notifications(pg_connection):
    pg_connection.notifies = []
    pg_connection.poll()
    assert 0 == len(pg_connection.notifies)


@pytest.mark.django_db(transaction=True)
def test_process_stored_notifications(pg_connection):
    Author.objects.create(name='Billy')
    Author.objects.create(name='Billy2')
    assert 2 == len(pg_connection.notifies)
    assert 2 == Notification.objects.count()
    assert 0 == Post.objects.count()
    _simulate_listener_does_not_receive_notifications(pg_connection)
    process_stored_notifications()
    pg_connection.poll()
    # One notification for each lockable channel
    assert 5 == len(pg_connection.notifies)
    process_notifications(pg_connection)
    assert 0 == Notification.objects.count()
    assert 2 == Post.objects.count()


@pytest.mark.django_db(transaction=True)
def test_recover_notifications(pg_connection):
    Author.objects.create(name='Billy')
    Author.objects.create(name='Billy2')
    pg_connection.poll()
    assert 2 == len(pg_connection.notifies)
    assert 2 == Notification.objects.count()
    assert 0 == Post.objects.count()
    _simulate_listener_does_not_receive_notifications(pg_connection)
    with patch('pgpubsub.listen.POLL', False):
        listen(recover=True)
    pg_connection.poll()
    assert 0 == Notification.objects.count()
    assert 2 == Post.objects.count()

@pytest.mark.django_db(transaction=True)
def test_recover_multiple_notifications(pg_connection):
    ENTITIES_COUNT = 5
    for i in range(ENTITIES_COUNT):
        Author.objects.create(name=f'Billy{i}')
    pg_connection.poll()
    assert ENTITIES_COUNT == len(pg_connection.notifies)
    assert ENTITIES_COUNT == Notification.objects.count()
    assert 0 == Post.objects.count()
    _simulate_listener_does_not_receive_notifications(pg_connection)
    with patch('pgpubsub.listen.POLL', False):
        listen(recover=True)
    pg_connection.poll()
    assert 0 == Notification.objects.count()
    assert ENTITIES_COUNT == Post.objects.count()


def _create_notification_that_cannot_be_processed():
    notification = Notification.objects.last()
    payload = json.loads(notification.payload)
    payload.pop('app', None)
    notification.payload = json.dumps(payload)
    notification.pk = None
    notification.save()


@pytest.mark.django_db(transaction=True)
def test_recover_notifications_after_exception(pg_connection):
    author = Author.objects.create(name='Billy')
    _create_notification_that_cannot_be_processed()
    Author.objects.create(name='Billy2')

    pg_connection.poll()
    assert 2 == len(pg_connection.notifies)

    assert 3 == Notification.objects.count()
    assert 0 == Post.objects.count()

    _simulate_listener_does_not_receive_notifications(pg_connection)
    with patch('pgpubsub.listen.POLL', False):
        listen(recover=True)
    pg_connection.poll()
    assert 1 == Notification.objects.count()
    assert 2 == Post.objects.count()

@pytest.mark.django_db(transaction=True)
def test_recover_multiple_notifications_after_exception(pg_connection):
    Author.objects.create(name=f'Billy_1')
    Author.objects.create(name=f'Billy_2')
    _create_notification_that_cannot_be_processed()
    Author.objects.create(name=f'Billy_3')
    _create_notification_that_cannot_be_processed()
    _create_notification_that_cannot_be_processed()
    _create_notification_that_cannot_be_processed()
    Author.objects.create(name=f'Billy_4')
    Author.objects.create(name=f'Billy_5')

    GOOD_COUNT = 5
    BROKEN_COUNT = 4

    pg_connection.poll()
    assert GOOD_COUNT == len(pg_connection.notifies)
    assert GOOD_COUNT + BROKEN_COUNT == Notification.objects.count()
    assert 0 == Post.objects.count()

    _simulate_listener_does_not_receive_notifications(pg_connection)
    with patch('pgpubsub.listen.POLL', False):
        listen(recover=True)
    pg_connection.poll()
    assert BROKEN_COUNT == Notification.objects.count()
    assert GOOD_COUNT == Post.objects.count()


@pytest.mark.django_db(transaction=True)
def test_media_insert_notify(pg_connection):
    Media.objects.create(name='avatar.jpg', content_type='image/png', size=15000)
    assert 1 == len(pg_connection.notifies)
    stored_notification = Notification.from_channel(channel=MediaTriggerChannel).get()
    assert 'old' in stored_notification.payload
    assert 'new' in stored_notification.payload


@pytest.fixture
def tx_start_time(django_db_setup):
    with connection.cursor() as cursor:
        cursor.execute("SELECT now();")
        return cursor.fetchone()[0]


@pytest.mark.django_db(transaction=True)
def test_persistent_notification_has_a_creation_timestamp(pg_connection, tx_start_time):
    Media.objects.create(name='avatar.jpg', content_type='image/png', size=15000)
    assert 1 == len(pg_connection.notifies)
    stored_notification = Notification.from_channel(channel=MediaTriggerChannel).get()
    assert stored_notification.created_at >= tx_start_time


@pytest.mark.django_db(transaction=True)
def test_persistent_notification_has_a_db_version(pg_connection, tx_start_time):
    latest_app_migration = MigrationRecorder.Migration.objects.filter(app='tests').latest('id')
    Media.objects.create(name='avatar.jpg', content_type='image/png', size=15000)
    assert 1 == len(pg_connection.notifies)
    stored_notification = Notification.from_channel(channel=MediaTriggerChannel).get()
    assert stored_notification.db_version == latest_app_migration.id



import datetime
import json
from dataclasses import dataclass
from decimal import Decimal

import pytest
from django.contrib.auth.models import User
from django.core.serializers.json import DjangoJSONEncoder
from django.utils import timezone

from pgpubsub import TriggerChannel
from pgpubsub.models import Notification
from pgpubsub.tests.channels import (
    AuthorTriggerChannel,
    ChildTriggerChannel,
    ChildOfAbstractTriggerChannel,
    MediaTriggerChannel,
    PostTriggerChannel,
)
from pgpubsub.tests.models import Post, Author, Child, ChildOfAbstract, Media


def test_deserialize_post_trigger_channel():
    @dataclass
    class MyChannel(TriggerChannel):
        model: Post

    some_datetime = datetime.datetime.utcnow()
    post = Post(content='some-content', date=some_datetime, pk=1, rating=Decimal("1.1"))

    deserialized = MyChannel.deserialize(
        json.dumps(
            {
                'app': 'tests',
                'model': 'Post',
                'old': None,
                'new': {
                    'content': 'some-content',
                    'date': some_datetime.isoformat(),
                    'id': post.pk,
                    # See https://github.com/Opus10/django-pgpubsub/issues/29
                    'old_field': 'foo',
                    'rating': Decimal("1.1"),
                },
            },
            cls=DjangoJSONEncoder,
        )
    )
    assert deserialized['new'].date == some_datetime
    assert deserialized['new'].content == post.content
    assert deserialized['new'].rating == post.rating
    assert deserialized['new'].author == post.author


@pytest.mark.django_db(transaction=True)
def test_deserialize_insert_payload():
    user = User.objects.create(username='Billy')
    media = Media.objects.create(
        name='avatar.jpg',
        content_type='image/png',
        size=15000,
    )
    author = Author.objects.create(
        name='Billy',
        user=user,
        alternative_name='Jimmy',
        profile_picture=media,
    )
    # Notification comes from the AuthorTriggerChannel
    # and contains a serialized version of the author
    # object in the payload attribute.
    insert_notification = Notification.from_channel(
        channel=AuthorTriggerChannel).get()
    deserialized = AuthorTriggerChannel.deserialize(
        insert_notification.payload)

    assert deserialized['new'].name == author.name
    assert deserialized['new'].alternative_name == author.alternative_name
    assert deserialized['new'].id == author.pk
    assert deserialized['new'].user == author.user
    assert deserialized['new'].profile_picture == author.profile_picture


@pytest.mark.django_db(transaction=True)
def test_deserialize_edit_payload():
    media = Media.objects.create(
        name='avatar.jpg',
        content_type='image/png',
        size=15000,
        store_id="some-value",
    )
    assert 1 == Notification.objects.all().count()
    insert_notification = Notification.from_channel(
        channel=MediaTriggerChannel).last()

    deserialized = MediaTriggerChannel.deserialize(
        insert_notification.payload)

    assert media.name == deserialized['new'].name
    assert media.pk == deserialized['new'].pk
    assert media.key == deserialized['new'].key
    assert media.size == deserialized['new'].size
    assert media.store_id == deserialized['new'].store_id

    media.name = 'avatar_2.jpg'
    media.save()

    assert 2 == Notification.objects.all().count()
    edit_notification = Notification.from_channel(
        channel=MediaTriggerChannel).last()

    deserialized = MediaTriggerChannel.deserialize(
        edit_notification.payload)

    assert deserialized['new'].name == media.name
    assert deserialized['new'].pk == media.pk
    assert deserialized['new'].key == media.key
    assert deserialized['new'].size == media.size
    assert media.store_id == deserialized['new'].store_id


@pytest.mark.django_db(transaction=True)
def test_deserialize_delete_payload():
    user = User.objects.create(username='Billy')
    author = Author.objects.create(name='Billy', user=user)

    post = Post.objects.create(
        author=author,
        content='my post',
        date=timezone.now(),
    )
    original_id = post.pk

    # When we delete a post, a notification is sent via
    # PostTriggerChannel
    post.delete()
    delete_notification = Notification.from_channel(
        channel=PostTriggerChannel).get()
    deserialized = PostTriggerChannel.deserialize(
        delete_notification.payload)

    assert deserialized['old'].author == post.author
    assert deserialized['old'].date.date() == post.date.date()
    assert deserialized['old'].date.time() == post.date.time()
    assert deserialized['old'].id == original_id
    assert deserialized['new'] is None


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize(
    "cls,trigger_channel_cls",
    [
        (Child, ChildTriggerChannel),
        (ChildOfAbstract, ChildOfAbstractTriggerChannel)
    ]
)
def test_deserialize_child_payload(cls, trigger_channel_cls):
    child = cls.objects.create()

    assert 1 == Notification.objects.all().count()
    insert_notification = Notification.from_channel(channel=trigger_channel_cls).last()

    deserialized = trigger_channel_cls.deserialize(insert_notification.payload)

    assert child.pk == deserialized['new'].pk
    assert child.key == deserialized['new'].key



# Generated by Django 3.2.12 on 2023-05-15 11:42

from django.db import migrations, models
from pgpubsub.models import Notification


class Migration(migrations.Migration):

    dependencies = [
        ('pgpubsub', '0002_notification_created_at'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='db_version',
            field=models.IntegerField(null=True),
        ),
        migrations.RunSQL(
            sql=f"""
                ALTER TABLE {Notification._meta.db_table}
                ADD CONSTRAINT pgpubsub_notification_db_version_fk
                FOREIGN KEY (db_version)
                REFERENCES django_migrations(id)
                ON DELETE RESTRICT
            """,
            reverse_sql=f"""
                ALTER TABLE {Notification._meta.db_table}
                DROP CONSTRAINT pgpubsub_notification_db_version_fk
            """,
        ),
    ]



from dataclasses import dataclass
import datetime
from typing import Dict, List, Set, Tuple

from pgpubsub.channel import Channel


def test_deserialize_1():
    @dataclass
    class MyChannel(Channel):
        arg1: str
        arg2: Dict[int, int]
        default_arg1: float = 0.0

    deserialized = _deserialize(MyChannel, arg1='1', arg2={1: 2}, default_arg1=3.4)
    assert {'arg1': '1', 'arg2': {1: 2},
            'default_arg1': 3.4} == deserialized


def test_deserialize_2():
    @dataclass
    class MyChannel(Channel):
        arg1: Dict[str, bool]
        default_arg1: bool = False
        default_arg2: int = 0

    deserialized = _deserialize(
        MyChannel, arg1={'Paul': False}, default_arg1=True)
    assert {'arg1': {'Paul': False},
            'default_arg1': True,
            'default_arg2': 0} == deserialized


def test_deserialize_3():
    @dataclass
    class MyChannel(Channel):
        arg1: datetime.date
        arg2: Dict[datetime.date, bool]
        arg3: Dict[str, datetime.datetime]

    deserialized = _deserialize(
        MyChannel,
        arg1=datetime.date(2021, 1, 1),
        arg2={
            datetime.date(2021, 1, 7): True,
            datetime.date(2021, 1, 17): False,
        },
        arg3={'chosen_date': datetime.datetime(2021, 1, 1, 9, 30)},
    )

    assert {
        'arg1': datetime.date(2021, 1, 1),
        'arg2': {datetime.date(2021, 1, 7): True, datetime.date(2021, 1, 17): False},
        'arg3': {'chosen_date': datetime.datetime(2021, 1, 1, 9, 30)},
    } == deserialized


def test_deserialize_4():
    @dataclass
    class MyChannel(Channel):
        arg1: List[datetime.date]
        arg2: Set[float]
        arg3: Tuple[str]

    deserialized = _deserialize(
        MyChannel,
        arg1=[datetime.date(2021, 1, 1), datetime.date(2021, 1, 2)],
        arg2={1.0, 2.1},
        arg3=('hello', 'world'),
    )
    assert {
        'arg1': [datetime.date(2021, 1, 1), datetime.date(2021, 1, 2)],
        'arg2': {1.0, 2.1},
        'arg3': ('hello', 'world'),
    } == deserialized


def _deserialize(channel_cls, **kwargs):
    serialized = channel_cls(**kwargs).serialize()
    return channel_cls.deserialize(serialized)



# Generated by Django 3.2.12 on 2023-05-16 11:51

from django.db import migrations
import pgtrigger.compiler
import pgtrigger.migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pgpubsub', '0003_notification_db_version'),
    ]

    operations = [
        pgtrigger.migrations.AddTrigger(
            model_name='notification',
            trigger=pgtrigger.compiler.Trigger(name='pgpubsub_notification_set_db_version', sql=pgtrigger.compiler.UpsertTriggerSql(func="\n                    NEW.db_version := (\n                        SELECT max(id)\n                        FROM django_migrations\n                        WHERE app = ((NEW.payload #>> '{}')::jsonb ->> 'app')\n                    );\n                    NEW.created_at := NOW();\n                    RETURN NEW;\n                ", hash='e04fb736d4f2f3925f3756bd31d0771aa16d4b6f', operation='INSERT', pgid='pgtrigger_pgpubsub_notification_set_db_version_ac4cd', table='pgpubsub_notification', when='BEFORE')),
        ),
    ]



# Generated by Django 3.2.12 on 2023-05-19 11:56

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0004_alter_media_id'),
    ]

    operations = [
        migrations.RenameField(
            model_name='media',
            old_name='id',
            new_name='key',
        ),
    ]



import hashlib
from abc import abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from decimal import Decimal
import datetime
import inspect
import json
from pydoc import locate
from typing import Callable, Dict, Union, List

from django.apps import apps
from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models


registry = defaultdict(list)


@dataclass
class BaseChannel:
    lock_notifications = False

    def __post_init__(self):
        self.callbacks = []

    @classmethod
    def name(cls):
        module_name = inspect.getmodule(cls).__name__
        return f'{module_name}.{cls.__name__}'

    @classmethod
    def listen_safe_name(cls):
        # Postgres LISTEN protocol accepts channel names
        # which are at most 63 characters long.
        model_hash = hashlib.sha1(
            cls.name().encode()).hexdigest()[:5]
        return f'pgpubsub_{model_hash}'

    @classmethod
    def get(cls, name: str):
        for channel_cls, callbacks in registry.items():
            if channel_cls.listen_safe_name() == name:
                return channel_cls, callbacks

    @classmethod
    def register(cls, callback: Callable):
        registry[cls].append(callback)

    @classmethod
    @abstractmethod
    def deserialize(cls, payload: Union[Dict, str]):
        if isinstance(payload, str):
            payload = json.loads(payload, parse_float=Decimal)
        return payload

    @classmethod
    def build_from_payload(
        cls,
        notification_payload: Union[Dict, str],
        callbacks: List[Callable],
    ):
        deserialized = cls.deserialize(notification_payload)
        channel = cls(**deserialized)
        channel.callbacks.extend(callbacks)
        return channel

    @property
    def signature(self):
        return {k: v for k, v in self.__dict__.items()
                if k in self.__dataclass_fields__}

    def execute_callbacks(self):
        for callback in self.callbacks:
            callback(**self.signature)


@dataclass
class Channel(BaseChannel):

    @classmethod
    def deserialize(cls, payload: Union[Dict, str]):
        payload = super().deserialize(payload)
        serialized_kwargs = payload['kwargs']
        kwargs = {}
        for kwarg_name, val in serialized_kwargs.items():
            deserialized_val = val
            kwarg_type = cls.__dataclass_fields__[kwarg_name].type
            origin_type = getattr(kwarg_type, '__origin__', kwarg_type)
            if origin_type is dict:
                key_type, val_type = kwarg_type.__args__
                deserialized_val = {
                    cls._deserialize_arg(key, key_type):
                        cls._deserialize_arg(val, val_type)
                    for key, val in val.items()
                }
            elif origin_type in (list, tuple, set):
                (element_type,) = kwarg_type.__args__
                deserialized_val = origin_type(
                    cls._deserialize_arg(x, element_type) for x in val)
            kwargs[kwarg_name] = cls._deserialize_arg(
                deserialized_val, origin_type)
        return kwargs

    def serialize(self):
        serialized_kwargs = {}
        for kwarg, val in self.signature.items():
            serialized_val = val
            kwarg_type = self.__dataclass_fields__[kwarg].type
            origin_type = getattr(kwarg_type, '__origin__', kwarg_type)
            if origin_type is dict:
                serialized_val = {
                    self._date_serial(k): self._date_serial(v)
                    for k, v in val.items()
                }
            elif origin_type in (list, tuple, set):
                serialized_val = [self._date_serial(x) for x in val]
            serialized_kwargs[kwarg] = serialized_val
        return json.dumps(
            {'kwargs': serialized_kwargs},
            default=self._date_serial,
            cls=DjangoJSONEncoder,
        )

    @staticmethod
    def _date_serial(obj):
        if isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()
        return obj

    @classmethod
    def _deserialize_arg(cls, arg, arg_type):
        if arg_type in (datetime.datetime, datetime.date):
            return arg_type.fromisoformat(arg)
        else:
            return arg_type(arg)


@dataclass
class TriggerChannel(BaseChannel):

    model = NotImplementedError
    old: models.Model
    new: models.Model

    @classmethod
    def deserialize(cls, payload: Union[Dict, str]):
        payload_dict = super().deserialize(payload)
        old_model_data = cls._build_model_serializer_data(payload_dict, state='old')
        new_model_data = cls._build_model_serializer_data(payload_dict, state='new')

        old_deserialized_objects = serializers.deserialize(
            'json',
            json.dumps(old_model_data, cls=DjangoJSONEncoder),
            ignorenonexistent=True,
        )
        new_deserialized_objects = serializers.deserialize(
            'json',
            json.dumps(new_model_data, cls=DjangoJSONEncoder),
            ignorenonexistent=True,
        )

        old = next(old_deserialized_objects, None)
        if old is not None:
            old = old.object
        new = next(new_deserialized_objects, None)
        if new is not None:
            new = new.object
        return {'old': old, 'new': new}

    @classmethod
    def _build_model_serializer_data(cls, payload: Dict, state: str):
        """Reformat serialized data into shape as expected
        by the Django model deserializer.
        """
        app = payload['app']
        model_name = payload['model']
        model_cls = apps.get_model(
            app_label=payload['app'],
            model_name=payload['model'],
        )
        fields = {
            field.name: field for field in model_cls._meta.fields
        }
        db_columns = {
            field.column: field for field in model_cls._meta.fields
        }

        original_state = payload[state]
        new_state = {}
        model_data = []
        if payload[state] is not None:
            for db_field in list(original_state):
                # Triggers serialize the notification payload with
                # respect to how the model fields look as columns
                # in the database. We therefore need to translate
                # to model fields and skip outdated fields
                value = original_state.pop(db_field)
                if db_field in db_columns:
                    model_field = db_columns[db_field].name
                    new_state[model_field] = value

            pk = model_cls._meta.pk
            serialized = {
                'fields': new_state,
                'pk': new_state[pk.name],
                'model': f'{app}.{model_name}',
            }
            if isinstance(pk, models.fields.related.OneToOneField):
                serialized['fields'][pk.remote_field.model._meta.pk.name] = serialized['pk']
            model_data.append(serialized)
        return model_data


def locate_channel(channel):
    if isinstance(channel, str):
        channel = locate(channel)
    if channel is None:
        raise ChannelNotFound()
    return channel


class ChannelNotFound(Exception):
    def __str__(self):
        return 'Channel not found!'



# Generated by Django 3.2.12 on 2023-05-19 14:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0007_child_pgpubsub_89ef9'),
    ]

    operations = [
        migrations.AlterField(
            model_name='parent',
            name='key',
            field=models.AutoField(editable=False, primary_key=True, serialize=False),
        ),
    ]



# Generated by Django 3.2.12 on 2022-04-07 10:27

from django.db import migrations, models

try:
    from django.db.models import JSONField
except:
    from django.contrib.postgres.fields import JSONField


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='Notification',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name='ID'
                    ),
                ),
                ('channel', models.CharField(max_length=63)),
                ('payload', JSONField()),
            ],
        ),
    ]



# Generated by Django 3.2.12 on 2023-05-22 07:20

from django.db import migrations, models
import pgtrigger.compiler
import pgtrigger.migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0008_alter_parent_key'),
    ]

    operations = [
        migrations.CreateModel(
            name='ChildOfAbstract',
            fields=[
                ('key', models.AutoField(editable=False, primary_key=True, serialize=False)),
            ],
            options={
                'abstract': False,
            },
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='childofabstract',
            trigger=pgtrigger.compiler.Trigger(name='pgpubsub_b1c0b', sql=pgtrigger.compiler.UpsertTriggerSql(declare='DECLARE payload TEXT;', func="\n            \n            payload := json_build_object(\n                'app', 'tests',\n                'model', 'ChildOfAbstract',\n                'old', row_to_json(OLD),\n                'new', row_to_json(NEW)\n              );\n        \n            \n            INSERT INTO pgpubsub_notification (channel, payload)\n            VALUES ('pgpubsub_b1c0b', to_json(payload::text));\n        \n            perform pg_notify('pgpubsub_b1c0b', payload);\n            RETURN NEW;\n        ", hash='392d173cf724ad909495797518008b67dcee4666', operation='UPDATE OR INSERT', pgid='pgtrigger_pgpubsub_b1c0b_c8531', table='tests_childofabstract', when='AFTER')),
        ),
    ]



# Generated by Django 3.2.12 on 2023-05-19 13:55

from django.db import migrations
import pgtrigger.compiler
import pgtrigger.migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0006_child_parent'),
    ]

    operations = [
        pgtrigger.migrations.AddTrigger(
            model_name='child',
            trigger=pgtrigger.compiler.Trigger(name='pgpubsub_89ef9', sql=pgtrigger.compiler.UpsertTriggerSql(declare='DECLARE payload TEXT;', func="\n            \n            payload := json_build_object(\n                'app', 'tests',\n                'model', 'Child',\n                'old', row_to_json(OLD),\n                'new', row_to_json(NEW)\n              );\n        \n            \n            INSERT INTO pgpubsub_notification (channel, payload)\n            VALUES ('pgpubsub_89ef9', to_json(payload::text));\n        \n            perform pg_notify('pgpubsub_89ef9', payload);\n            RETURN NEW;\n        ", hash='d2d276ef8c024dff4f84c145a99fabb8890d7b8a', operation='UPDATE OR INSERT', pgid='pgtrigger_pgpubsub_89ef9_92bc1', table='tests_child', when='AFTER')),
        ),
    ]



# Generated by Django 3.2.12 on 2023-05-19 13:50

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0005_rename_id_media_key'),
    ]

    operations = [
        migrations.CreateModel(
            name='Parent',
            fields=[
                ('key', models.AutoField(primary_key=True, serialize=False)),
            ],
        ),
        migrations.CreateModel(
            name='Child',
            fields=[
                ('parent_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='tests.parent')),
            ],
            bases=('tests.parent',),
        ),
    ]



# Generated by Django 3.2.12 on 2023-04-11 08:29

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='author',
            name='alternative_name',
            field=models.TextField(db_column='other', null=True),
        ),
        migrations.AlterField(
            model_name='author',
            name='profile_picture',
            field=models.ForeignKey(db_column='picture', null=True, on_delete=django.db.models.deletion.PROTECT, to='tests.media'),
        ),
    ]
